#pragma once 

class AnimatedCharacter;
class AnimatedCharacterInstance;
class CBaseCamera;
class NVUTMesh;

#include "nvutskybox.h"

#define INTERNAL_MAX_INSTANCES_PER_BUFFER 682
#define INTERNAL_MAX_BUFFERS 14

/*
    The Army Manager wraps up the full effect.  It handles loading the animation data, creation of all DX10 resources, and rendering.
    Important bits:
    - AnimatedCharacter contains the mesh data, broken into seperate meshes to allow instancing of each piece.
    - Characters are rendered with instancing, and the custom data per instance is stored in a buffer in each instance draw call struct
    - All animation frames are stored in a texture resource, m_pAnimationTexture per LOD
    - Color maps for all polys are set for each draw, 
*/
class ArmyManager
{
public:
    static ArmyManager *singleton()
    {
        if(m_Singleton==NULL) m_Singleton = new ArmyManager();
        return m_Singleton;
    }

    static void destroy()
    {
        delete m_Singleton;
    }

    struct InstanceDataElement
    {
        D3DXVECTOR4 world1;            // the world transform for this matrix row 1
        D3DXVECTOR4 world2;            // the world transform for this matrix row 2
        D3DXVECTOR4 world3;            // the world transform for this matrix row 3 (row 4 is implicit)
        D3DXCOLOR color;

        // $ Technically this is bundled, but there is not class that makes a uint vector, so just keep flat
        UINT animationIndex;            // offset in vectors into the whole data stream for the start of the animation playing
        UINT frameOffset;            // offset in vectors into the animation stream for the start of the frame playing
        UINT attachmentSet;            // the id to determine which geo attachments get set
        UINT lerpValue;            // lerp between frames
    };

    struct InstancedDrawCallState
    {
        InstancedDrawCallState() 
        {
            for(int i=0;i<INTERNAL_MAX_BUFFERS;i++)
                pInstanceDataBuffer[i] = NULL;numToDraw=0;
        }
        ID3D10Buffer *pInstanceDataBuffer[INTERNAL_MAX_BUFFERS];
        std::vector<int> instancesUsingThisMesh;
        int numToDraw;  // number to draw form this batch
        AnimatedCharacter *referenceCharacter;
        int meshIndex;
        int lod;
    };

    HRESULT Initialize(ID3D10Device* pd3dDevice, int bbWdith, int bbHeight);
    HRESULT Update(float deltaTime,const D3DXVECTOR3 &cameraAt, const D3DXVECTOR3 &camDir);
    HRESULT Render(ID3D10Device* pd3dDevice,D3DXMATRIX &mView,D3DXMATRIX &mProj, float time);
    void Release();
    HRESULT UpdateInstancedDrawState(ID3D10Device *pd3dDevice, bool bUpdateBuffers=true);
    void FillDataElement(InstanceDataElement*pData,int character);
    void ResizedSwapChain(ID3D10Device *pd3dDevice, int bbWdith, int bbHeight);

    void SetViewportSize(int w,int h);
    void SetEffectFloat(LPCSTR name, float value) {m_pEffect10->GetVariableByName(name)->AsScalar()->SetFloat(value); }
    void SetLODParams(float lod1, float lod2, float lod3);
    void SetHDR(bool bHDR){m_bHDR = bHDR;}
    bool GetHDR(){return m_bHDR;}
    void SetShadows(bool bShadows){m_bShadows=bShadows;}
    void SetInstancingMode(int mode){m_iInstancingMode = mode;}
    void SetVisualizeLOD(bool vis){m_bVisualizeLOD=vis;}
    bool GetVisualizeLOD(){return m_bVisualizeLOD;}
    void SetDisableUpdates(bool disable){m_bDisableUpdates=disable;}

    HRESULT CreateInstanceBuffer( ID3D10Device* pd3dDevice, ID3D10Buffer **pBuffer, int maxCharacters );
    HRESULT CreateAnimationTexture(ID3D10Device* pd3dDevice,AnimatedCharacter *pAnimatedCharacter, ID3D10Texture2D **ppAnimationTexture, ID3D10ShaderResourceView **ppAnimationResourceView);

    UINT GetAnimationOffset(AnimatedCharacter *pAnimatedCharacter,int animation);
    void SetNumCharacters(int num);
    int GetMaxCharacters() {return (INTERNAL_MAX_BUFFERS * INTERNAL_MAX_INSTANCES_PER_BUFFER)-1;}

    HRESULT Draw(ID3D10Device *pd3dDevice,D3DXMATRIX &mView,D3DXMATRIX &mProj, float time, int maxLOD=3);
    HRESULT DrawSorted(ID3D10Device *pd3dDevice,D3DXMATRIX &mView,D3DXMATRIX &mProj, float time, int maxLOD=3);
    HRESULT DrawHybridInstanced(ID3D10Device *pd3dDevice,D3DXMATRIX &mView,D3DXMATRIX &mProj, float time, int maxLOD=3);

    // Game logic
    D3DXVECTOR3 GetRandomScenePosition();
    int GetPolysDrawn() {return m_polysDrawnLastFrame;}
    int GetDrawCalls() {return m_drawCallsLastFrame;}

    // D3D10 Rendering vars (Release needed)
    ID3D10InputLayout*          m_pInputLayout;
    ID3D10InputLayout*          m_pInputLayoutTerrain;
    ID3D10Effect*               m_pEffect10;

    ID3D10Texture2D*            m_pAnimationTexture[4];
    ID3D10ShaderResourceView*   m_pAnimationResourceView[4];

    ID3D10Texture2D *           m_pHDRRT;
    ID3D10ShaderResourceView *  m_pHDRRTSRV;
    ID3D10RenderTargetView *    m_pHDRRTRTV;
    ID3D10Texture2D *           m_pHDRRT2;
    ID3D10ShaderResourceView *  m_pHDRRT2SRV;
    ID3D10RenderTargetView *    m_pHDRRT2RTV;
    ID3D10Texture2D *           m_pHDRRT3;
    ID3D10ShaderResourceView *  m_pHDRRT3SRV;
    ID3D10RenderTargetView *    m_pHDRRT3RTV;
    ID3D10Texture2D *           m_pHDRDepth;
    ID3D10DepthStencilView *    m_pHDRDepthDSV;

    ID3D10Texture2D *           m_pSMRT;
    ID3D10ShaderResourceView *  m_pSMRTSRV;
    ID3D10RenderTargetView *    m_pSMRTRTV;

    NVUTMesh                    m_GroundMesh;
    AnimatedCharacter *         m_CharacterLODs[4];

    NVUTSkybox                  m_skybox;

    std::vector<InstancedDrawCallState> m_InstancedDraws;
    
    // No need to Release these vars
    AnimatedCharacter *         m_ReferenceCharacter;   // no release needed since it is also in m_CharacterLODs[]
    ID3D10EffectTechnique*      m_pTechAnimationInstanced[3];
    ID3D10EffectTechnique*      m_pTechAnimation[3];
    ID3D10EffectTechnique*      m_pTechAnimInstAO;
    ID3D10EffectTechnique*      m_pTechTerrain;
    ID3D10EffectTechnique*      m_pTechHDR;

    ID3D10EffectMatrixVariable*             m_pWorldVar;
    ID3D10EffectMatrixVariable*             m_pWorldITVar;
    ID3D10EffectMatrixVariable*             m_pViewProjVar;
    ID3D10EffectMatrixVariable*             m_pWorldViewProjectionVar;
    ID3D10EffectVectorVariable*             m_pViewportSizeVar;
    ID3D10EffectShaderResourceVariable*     m_pAnimationsVar;
    ID3D10EffectShaderResourceVariable*     m_pDiffuseTexVar;
    ID3D10EffectShaderResourceVariable*     m_pNormalTexVar;
    ID3D10EffectShaderResourceVariable*     m_pEnvMapTexVar;

    // Character data vars
    bool m_bVisualizeLOD;
    bool m_bHDR;
    bool m_bShadows;
    bool m_bDisableUpdates;
    int m_iInstancingMode;
    int m_MaxCharacters;
    int m_NumToDraw;

    std::vector<AnimatedCharacterInstance*> m_Characters;

    int m_polysDrawnLastFrame;
    int m_drawCallsLastFrame;

protected:

    InstanceDataElement* m_pDataTemp;

private:
    ArmyManager();
    static ArmyManager*m_Singleton;

};  