//--------------------------------------------------------------------------------------
// This is the shader code for the SkinnedInstancing sample.
//--------------------------------------------------------------------------------------

#ifndef MATRIX_PALETTE_SIZE_DEFAULT
#define MATRIX_PALETTE_SIZE_DEFAULT 50
#endif

#ifndef MAX_INSTANCE_CONSTANTS
#define MAX_INSTANCE_CONSTANTS 682
#endif

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------

float4x4    g_mWorld : World;                  // World matrix for object
float4x4    g_mWorldIT : WorldIT;                  // World IT matrix for object
float4x4    g_mViewProj : ViewProjection;
float4x4    g_mWorldViewProjection : WorldViewProjection;    // World * View * Projection matrix
float       g_TexOffset : TEXCOORDANIMFACTOR;
float3      g_EyePos : CAMERAPOS;

float2      g_viewportSize : ViewportSizePixels;
float4x4    g_ShadowMatrix : SHADOWTRANSFORM;
bool        g_bShadow = true;

cbuffer animationvars
{
    float3x4    g_matrices[MATRIX_PALETTE_SIZE_DEFAULT];
    float2      g_UVOffset = float2(0,0);
    float3      g_instanceColor = float3(1,1,1);
}

struct PerInstanceData
{
    float4 world1;
    float4 world2;
    float4 world3;
    float4 color;
    uint4    animationData;
};

cbuffer cInstanceData
{
    PerInstanceData    g_Instances[MAX_INSTANCE_CONSTANTS];
}

cbuffer config
{
    uint g_InstanceMatricesWidth;
};

cbuffer cLighting
{
    float3      g_lightPos = float3(0,500,0);
    float3      g_lightColor = (0.85,0.85,1.0);
    float4x4    g_Identity = {    1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};
    float3      g_fogColor = float3(0.6,0.6,0.7);
    float       g_Zero = 0;
    
    float       Kd = 1.9;
    float       Ks = 0.9;
    float       Kb = 0.85;
    float       Kr = 0.5;
    float       Kf = 0.2;
    float       SpecExpon = 64.0;
    float       AmbientScale = 0.8;
    float4      AmbiColor = float4(0.5,0.5,0.5,0);
    float       Exposure = 1;
    float       Gamma = 1;
};


//--------------------------------------------------------------------------------------
// RenderStates
//--------------------------------------------------------------------------------------
RasterizerState Wire
{
    FillMode=Wireframe;
    CullMode=None;
};

RasterizerState NoCull
{
	CullMode = None;
};

RasterizerState Cull
{
	CullMode = Back;
};

BlendState Opaque
{
    BlendEnable[0] = false;
};

BlendState ShadowMapWrite
{
    RenderTargetWriteMask[0] = 0x1;
    RenderTargetWriteMask[1] = 0x0;
    RenderTargetWriteMask[2] = 0x0;
    RenderTargetWriteMask[3] = 0x0;
    RenderTargetWriteMask[4] = 0x0;
    RenderTargetWriteMask[5] = 0x0;
    RenderTargetWriteMask[6] = 0x0;
    RenderTargetWriteMask[7] = 0x0;
};

DepthStencilState DisableDepthTestWrite
{
    DepthEnable = FALSE;
    DepthWriteMask = 0;
};

DepthStencilState EnableDepthTestWrite
{
    DepthEnable = TRUE;
    DepthWriteMask = ALL;
};


//--------------------------------------------------------------------------------------
// Texture samplers
//--------------------------------------------------------------------------------------
Texture2D g_txDiffuse       : DIFFUSE;
Texture2D g_txNormals       : NORMAL;
TextureCube g_txEnvMap      : ENVMAP;   // the environmap

Texture2D g_txAnimations    : ANIMATIONS;

Texture2D g_txAOPositions     : POSITIONS;    // the position of the geometry rendered to each screen pixel
Texture2D g_txAONormals     : NORMALS;    // the normal of the geometry rendered to each screen pixel
Texture2D g_txAOColors     : COLORS;    // the color of the geometry rendered to each screen pixel

Texture2D g_HDR : HDR;  // HDR rendering of full scene
Texture2D g_HDRBlur : HDRBLUR;  // HDR rendering of full scene
Texture2D g_ShadowMap : SHADOWMAP;  // contains a full scene shadow map

SamplerState g_samLinear
{
    Filter = MIN_MAG_MIP_LINEAR;
    AddressU = Wrap;
    AddressV = Wrap;
};
SamplerState g_samAniso
{
    Filter = ANISOTROPIC;
    AddressU = Wrap;
    AddressV = Wrap;
    MAXLOD=5;
};
SamplerState g_samPoint
{
    Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
};

//--------------------------------------------------------------------------------------
// Vertex shader output structure
//--------------------------------------------------------------------------------------

struct A_to_VS
{
    float4 vPos         : POSITION;
    float3 vNormal      : NORMAL;
    float3 vTangent     : TANGENT;
    float2 vTexCoord0   : TEXCOORD0;
    float4 vBones       : BONES;
    float4 vWeights     : WEIGHTS;
    uint   InstanceId   : SV_InstanceID;
};

struct VS_to_PS
{
    float4 pos      : SV_Position;
    float3 color    : COLOR;
    float3 tex      : TEXTURE0;
    float3 norm     : NORMAL;
    float3 tangent  : TANGENT;
    float3 binorm   : BINORMAL;
    float3 worldPos : WORLDPOS;
};

struct VS_to_PS_Low
{
    float4 pos      : SV_Position;
    float3 color    : COLOR;
    float3 tex      : TEXTURE0;
    float3 norm     : NORMAL;
    float3 worldPos : WORLDPOS;    
};

struct VS_to_PS_VertLighting
{
    float4 pos      : SV_Position;
    float4 color    : COLOR;    // color and w is diffuse coefficient
    float3 tex      : TEXTURE0;
};

struct VS_to_PS_FSQuad
{
    float4 pos      : SV_Position;
    float2 tex      : TEXTURE0;
};

// Now include our utility FXH, MUST be after all the struct defs, etc.
#include "Lighting.fxh"

float4x4 decodeMatrix(float3x4 encodedMatrix)
{
    return float4x4(    float4(encodedMatrix[0].xyz,0),
                        float4(encodedMatrix[1].xyz,0),
                        float4(encodedMatrix[2].xyz,0),
                        float4(encodedMatrix[0].w,encodedMatrix[1].w,encodedMatrix[2].w,1));
}

// Read a matrix(3 texture reads) from a texture containing animation data
float4x4 loadBoneMatrix(uint3 animationData,float bone)
{
    // Calculate a UV for the bone for this vertex
    float4x4 rval = g_Identity;
    
    // animationData.x and .y are linear offsets.  Combine into a single linear offset.
    uint baseIndex = animationData.x + animationData.y;
    baseIndex += (4*bone);    // 4*bone is since each bone is 4 texels to form a float4x4 
        
    // Now turn linear offset into 2D coords
    uint baseU = baseIndex%g_InstanceMatricesWidth;
    uint baseV = baseIndex/g_InstanceMatricesWidth;
   
    // Note that we assume the width of the texture(and just add texels) is an even multiple of the # of texels per bone,
    //     otherwise we'd have to recalculate the V component per lookup
    float4 mat1 = g_txAnimations.Load( uint3(baseU,baseV,0));
    float4 mat2 = g_txAnimations.Load( uint3(baseU+1,baseV,0));
    float4 mat3 = g_txAnimations.Load( uint3(baseU+2,baseV,0));
      
    // only load 3 of the 4 values, and deocde the matrix from them.
    rval = decodeMatrix(float3x4(mat1,mat2,mat3));
    
    return rval;
}

VS_to_PS CharacterAnimatedInstancedVS( A_to_VS input )
{
    VS_to_PS output;
    float3 vNormalWorldSpace;
    
    uint4 animationData = g_Instances[input.InstanceId].animationData;    
    
       // Our per instance data is stored in constants
    float4 worldMatrix1 = g_Instances[input.InstanceId].world1;
    float4 worldMatrix2 = g_Instances[input.InstanceId].world2;
    float4 worldMatrix3 = g_Instances[input.InstanceId].world3;
    float4 instanceColor = g_Instances[input.InstanceId].color;
    
    float4x4 finalMatrix;
    finalMatrix = input.vWeights.x * loadBoneMatrix(animationData,input.vBones.x);

    if(input.vWeights.y > 0) 
    {
        finalMatrix += input.vWeights.y * loadBoneMatrix(animationData,input.vBones.y);
        if(input.vWeights.z > 0)
        {
            finalMatrix += input.vWeights.z * loadBoneMatrix(animationData,input.vBones.z);
            if(input.vWeights.w > 0) finalMatrix += input.vWeights.w * loadBoneMatrix(animationData,input.vBones.w);    
        }
    }
    
    float4 vAnimatedPos     = mul(float4(input.vPos.xyz,1),finalMatrix);
    float4 vAnimatedNormal  = mul(float4(input.vNormal.xyz,0),finalMatrix);
    float4 vAnimatedTangent = mul(float4(input.vTangent.xyz,0),finalMatrix);

    // the whole model position, decode a little, we have compressed for space.
    float4x4 worldMatrix = decodeMatrix(float3x4(worldMatrix1,worldMatrix2,worldMatrix3));
    float4 vWorldAnimatedPos = mul(vAnimatedPos, worldMatrix);

    // Transform the position from object space to homogeneous projection space
    output.pos = mul(vWorldAnimatedPos, g_mWorldViewProjection);

    // Transform the normal from object space to world space    
    output.norm = normalize(mul(vAnimatedNormal.xyz,(float3x3)worldMatrix ));
    output.tangent = normalize(mul(vAnimatedTangent.xyz, (float3x3)worldMatrix ));
    output.binorm = cross(output.norm,output.tangent);
    
    // Do the position too for lighting
    output.worldPos = float4(vWorldAnimatedPos.xyz,1);
    output.tex.xy = float2(input.vTexCoord0.x,-input.vTexCoord0.y); 
    output.tex.z = 0;
    
    output.color = instanceColor.xyz;
    
    return output;
}

VS_to_PS CharacterInstancedVS( A_to_VS input ) 
{
    VS_to_PS output;
    float3 vNormalWorldSpace;
    
       // Our per instance data is stored in constants
    float4 worldMatrix1 = g_Instances[input.InstanceId].world1;
    float4 worldMatrix2 = g_Instances[input.InstanceId].world2;
    float4 worldMatrix3 = g_Instances[input.InstanceId].world3;
    float4 instanceColor = g_Instances[input.InstanceId].color;
    
    float4 vAnimatedPos = float4(input.vPos.xyz,1);
    float4 vAnimatedNormal = float4(input.vNormal.xyz,0);
    float4 vAnimatedTangent = float4(input.vTangent.xyz,0);
    
    // the whole model position, decode a little, we have compressed for space.
    float4x4 worldMatrix = decodeMatrix(float3x4(worldMatrix1,worldMatrix2,worldMatrix3));
    vAnimatedPos = mul(vAnimatedPos, worldMatrix );
    vAnimatedNormal = mul(float4(vAnimatedNormal.xyz,0), worldMatrix );
    vAnimatedTangent = mul(float4(vAnimatedTangent.xyz,0), worldMatrix );
    
    // Transform the position from object space to homogeneous projection space
    output.pos = mul(vAnimatedPos, g_mWorldViewProjection);

    // Transform the normal from object space to world space    
    output.norm = normalize(vAnimatedNormal.xyz);
    output.tangent = normalize(vAnimatedTangent.xyz);
    output.binorm = cross(output.norm,output.tangent);
    
    // Do the position too for lighting
    output.worldPos = float4(vAnimatedPos.xyz,1);
    output.tex.xy = float2(input.vTexCoord0.x,-input.vTexCoord0.y); 
    output.tex.z = 0;
    
    output.color = instanceColor.xyz;
    
    return output;
}


VS_to_PS CharacterAnimatedVS( A_to_VS input )
{
    VS_to_PS output;
    float3 vNormalWorldSpace;
                
    float4x4 finalMatrix;
    finalMatrix = input.vWeights.x * decodeMatrix(g_matrices[input.vBones.x]);

    if(input.vWeights.y > 0) 
    {
        finalMatrix += input.vWeights.y * decodeMatrix(g_matrices[input.vBones.y]);
        if(input.vWeights.z > 0)
        {
            finalMatrix += input.vWeights.z * decodeMatrix(g_matrices[input.vBones.z]);
            if(input.vWeights.w > 0) finalMatrix += input.vWeights.w * decodeMatrix(g_matrices[input.vBones.w]);    
        }
    }
    
    float4 vAnimatedPos = mul(float4(input.vPos.xyz,1),finalMatrix);
    float4 vAnimatedNormal = mul(float4(input.vNormal.xyz,0),finalMatrix);
    float4 vAnimatedTangent = mul(float4(input.vTangent.xyz,0),finalMatrix);
        
    // Transform the position from object space to homogeneous projection space
    output.pos = mul(vAnimatedPos, g_mWorldViewProjection);

    // Transform the normal from object space to world space    
    output.norm = normalize(mul(vAnimatedNormal.xyz, (float3x3)g_mWorld)); // normal to world space
    output.tangent = normalize(mul(vAnimatedTangent.xyz, (float3x3)g_mWorld));
    output.binorm = cross(output.norm,output.tangent);
    
    // Do the position too for lighting
    float4 vWorldAnimatedPos = mul(float4(vAnimatedPos.xyz,1), g_mWorld);
    output.worldPos = float4(vWorldAnimatedPos.xyz,1);
    output.tex.xy = float2(input.vTexCoord0.x,-input.vTexCoord0.y); 
    output.tex.z = 0;
    
    output.color = g_instanceColor;
    
    return output;    
}

VS_to_PS CharacterVS( A_to_VS input )
{
    VS_to_PS output;
    float3 vNormalWorldSpace;

    float4 vAnimatedPos = float4(input.vPos.xyz,1);
    float4 vAnimatedNormal = float4(input.vNormal.xyz,0);
    float4 vAnimatedTangent = float4(input.vTangent.xyz,0);
        
    // Transform the position from object space to homogeneous projection space
    output.pos = mul(vAnimatedPos, g_mWorldViewProjection);

    // Transform the normal from object space to world space    
    output.norm = normalize(mul(vAnimatedNormal.xyz, (float3x3)g_mWorld)); // normal to world space
    output.tangent = normalize(mul(vAnimatedTangent.xyz, (float3x3)g_mWorld));
    output.binorm = cross(output.norm,output.tangent);
    
    // Do the position too for lighting
    output.worldPos = mul(float4(vAnimatedPos.xyz,1), g_mWorld);
    output.tex.xy = float2(input.vTexCoord0.x,-input.vTexCoord0.y); 
    output.tex.z = 0;
    
    output.color = g_instanceColor;
    
    return output;    
}

VS_to_PS TerrainVS( float4 vPos : POSITION,
                         float3 vNormal : NORMAL,
                         float2 vTexCoord0 : TEXCOORD0,
                         float3 vTangent : TANGENT)
{
    VS_to_PS output;
    
    // Transform the position from object space to homogeneous projection space
    output.pos = mul(float4(vPos.xyz,1), g_mWorldViewProjection);

    // Transform the normal from object space to world space    
    output.norm = (mul(vNormal, (float3x3)g_mWorld)); // normal (world space)
    
    // $$ Mesh isn't generating tangents properly, so we are doing this in the shader.  REMOVE ME AND FIX THE MESH!!
    output.tangent = (mul(vNormal.zxy, (float3x3)g_mWorld)); // tangent (world space)
    output.binorm = cross(output.norm,output.tangent);
    
//    output.tangent = normalize(mul(vTangent, (float3x3)g_mWorld));
//    output.binorm = cross(output.norm,output.tangent);
    
    // Do the position too for lighting
    output.worldPos = mul(float4(vPos.xyz,1), g_mWorld);
    output.tex.xy = float2(vTexCoord0.x,vTexCoord0.y); 
    
    output.color = float4(1,1,1,1);

    return output;    
}

VS_to_PS_FSQuad FSQuadVS(uint vertexID: SV_VertexID)
{
	VS_to_PS_FSQuad output;
	
	if(vertexID == 0)
	{
	    output.pos = float4(-1,1,0,1);
	    output.tex = float2(0,0);
	}
	else if(vertexID == 1)
	{
        output.pos = float4(-1,-1,0,1);	
	    output.tex = float2(0,1);        
    }
    else if(vertexID == 2)
	{
        output.pos = float4(1,1,0,1);	
	    output.tex = float2(1,0);	
	}
    else if(vertexID == 3)
	{
        output.pos = float4(1,-1,0,1);	
	    output.tex = float2(1,1);        
    }
   
	return output;
}

//--------------------------------------------------------------------------------------

float4 CharacterPS( VS_to_PS input) : SV_Target
{
    float4 color = g_txDiffuse.Sample(g_samAniso,input.tex.xyz);
    //color.w = 0;
    color = color * float4(lerp(float3(1.0,1.0,1.0),input.color.xyz,color.w),1);
    
    float4 bumps = g_txNormals.Sample(g_samAniso,input.tex.xyz);    
    float gloss = bumps.w;
    bumps = float4(2*bumps.xyz - float3(1,1,1),0);
    
    float3 lightVec = normalize(g_lightPos - input.worldPos);    
    
    // lighting
    float3 Nn = normalize(input.norm);
    float3 Tn = normalize(input.tangent);
    float3 Bn = normalize(input.binorm);
    
    float3x3 TangentToWorld = float3x3(Tn,Bn,Nn);
    float3 Nb = lerp(Nn,mul(bumps.xyz,TangentToWorld),Kb);
    
    // override normal with normal map value
    Nb = normalize(Nb);
    input.norm = Nb;
    
    float3 diffContrib;
    float3 specContrib;
    float3 reflColor;
	shared_lighting_calculations(input,diffContrib,specContrib,reflColor);
	float3 Vn = normalize(g_EyePos - input.worldPos.xyz);
	
	float3 actualDiffuse = lerp(color*Kd*diffContrib,reflColor,Kr*gloss*fresnel(Vn,Nb,Kf));
	float3 result = actualDiffuse + gloss*Ks*specContrib;
	
	// finally lerp out the color on the edges so called rim lighting
	float rim = pow(1-saturate(dot(Vn,Nn)),2);
	if(rim < 0.5) rim = 0;
	
	//return float4(Vn,1);
	
	result = lerp(result,float4(0,0,0,0),rim);
    return float4(2*result,1);
}

float4 CharacterMidLODPS( VS_to_PS input) : SV_Target
{
    float4 color = g_txDiffuse.Sample(g_samAniso,input.tex.xyz);
    //color.w = 0;
    color = color * float4(lerp(float3(1.0,1.0,1.0),input.color.xyz,color.w),1);
    
    float gloss = 0.3;
    float3 lightVec = normalize(g_lightPos - input.worldPos);    
    
    float3 diffContrib;
    float3 specContrib;
    float3 reflColor;
    
	shared_lighting_calculations(input,diffContrib,specContrib,reflColor);
	
	float3 Vn = normalize(g_EyePos - input.worldPos.xyz);
	float3 Nn = normalize(input.norm);
	
	float3 result = lerp(color*Kd*diffContrib,reflColor*Kr,gloss*fresnel(Vn,Nn,Kf)) + gloss*Ks*specContrib;
	
	// finally lerp out the color on the edges so called rim lighting
	float rim = pow(1-saturate(dot(Vn,Nn)),2);
	if(rim < 0.7) rim = 0;
	
	result = lerp(result,float4(0,0,0,0),rim);

	
    return float4(2*result,1);

}

float4 CharacterLowLODPS( VS_to_PS input) : SV_Target
{
    float4 color = g_txDiffuse.Sample(g_samAniso,input.tex.xyz);
    if(color.w < 0.01) discard;    
    
    color.w = 1-color.w;
    
    color = color * 1.2*float4(lerp(float3(1.0,1.0,1.0),input.color.xyz,color.w),1);
    
    return float4(2*color.xyz,0);
    //return float4(color.wwww);
}

float Kd_t = 0.4f;
float4 AmbiColor_t = float4(0.9,0.9,0.9,0.9);
float4 TerrainPS( VS_to_PS input) : SV_Target
{
    float4 color = g_txDiffuse.Sample(g_samAniso,input.tex.xyz);
    float4 bumps = g_txNormals.Sample(g_samAniso,input.tex.xyz);    
    bumps = float4(2*bumps.xyz - float3(1,1,1),0); 
    
    float3 lightVec = normalize(g_lightPos - input.worldPos);    
    
    // lighting
    float3 Nn = normalize(input.norm);
    float3 Tn = normalize(input.tangent);
    float3 Bn = normalize(input.binorm);

    float3x3 TangentToWorld = float3x3(Tn,Bn,Nn);
    float3 Nb = lerp(Nn,mul(bumps.xyz,TangentToWorld),Kb);
        
    Nb = normalize(Nb);
        
    float3 diffContrib;
    float3 specContrib;
    
    float3 Ln = normalize(g_lightPos - input.worldPos);    
    float3 Vn = normalize(g_EyePos - input.worldPos.xyz);
    
    float lightFactor = dot(Ln,Nb);
    diffContrib = dot(Ln,Nb) * g_lightColor + AmbiColor;
    
    float3 result = color*Kd_t*diffContrib;
    
    // Silly simple shadows
    float shadowMapValue = 1.0;    
    if(g_bShadow && Nn.y > 0.9)
    {
        float3 shadowSpacePos = mul(float4(input.worldPos.xyz,1),g_ShadowMatrix).xyz;
        shadowSpacePos.xy *= float2(0.5,-0.5);
        shadowSpacePos.xy += float2(0.5,0.5);
        
        float2 smTexel = float2(0.5/1024.0,0.5/2048.0);
        shadowMapValue = shadowSpacePos.z>g_ShadowMap.Sample(g_samLinear,shadowSpacePos.xy).x?0.0:1.0;
        if(shadowMapValue < 1.0)
        {
            shadowMapValue += shadowSpacePos.z>g_ShadowMap.Sample(g_samLinear,shadowSpacePos.xy+float2(smTexel.x,smTexel.y)).x?0.0:1.0;
            shadowMapValue += shadowSpacePos.z>g_ShadowMap.Sample(g_samLinear,shadowSpacePos.xy+float2(-smTexel.x,smTexel.y)).x?0.0:1.0;
            shadowMapValue += shadowSpacePos.z>g_ShadowMap.Sample(g_samLinear,shadowSpacePos.xy+float2(smTexel.x,-smTexel.y)).x?0.0:1.0;
            shadowMapValue += shadowSpacePos.z>g_ShadowMap.Sample(g_samLinear,shadowSpacePos.xy-float2(-smTexel.x,-smTexel.y)).x?0.0:1.0;
            shadowMapValue /= 10;
            
            shadowMapValue += 0.5;            
            
            //shadowMapValue *= 2;
            //shadowMapValue *= dot(Nn,float3(0,1,0));
        }
    }
    
    return float4(shadowMapValue*result,1);
    //return float4(input.tangent.xyz,0);
}

float WriteDepthValue( VS_to_PS input) : SV_Target
{
    return input.pos.z;
}

float4 ExposureGammaPS( VS_to_PS_FSQuad input) : SV_Target
{
    float4 source = g_HDR.Sample(g_samAniso,input.tex.xy);
    float4 blur = 0.5*g_HDRBlur.Sample(g_samAniso,input.tex.xy);

    source = source + blur;
    
    // do adjustments on luminance...
    float sourceLuminance = 0.2126*source.x + 0.7152*source.y + 0.0722*source.z;
    
    float newLuminance = sourceLuminance * Exposure;
    newLuminance = pow(newLuminance,Gamma);
    
    source *= (newLuminance/sourceLuminance);
    return source;
}

#define WT9_0 1.0
#define WT9_1 0.8
#define WT9_2 0.6
#define WT9_3 0.4
#define WT9_4 0.2
#define WT9_NORMALIZE (WT9_0+2.0*(WT9_1+WT9_2+WT9_3+WT9_4))
float4 BlurHPS( VS_to_PS_FSQuad input) : SV_Target
{
    float4 source;
    float2 texelIncrement = float2(1.0 / g_viewportSize.x,0);
    source  = g_HDR.Sample(g_samAniso,input.tex.xy +   texelIncrement) * (WT9_1/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy + 2*texelIncrement) * (WT9_2/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy + 3*texelIncrement) * (WT9_3/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy + 4*texelIncrement) * (WT9_4/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy                   ) * (WT9_0/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy -   texelIncrement) * (WT9_1/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy - 2*texelIncrement) * (WT9_2/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy - 3*texelIncrement) * (WT9_3/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy - 4*texelIncrement) * (WT9_4/WT9_NORMALIZE);
    return source;
}

float4 BlurVPS( VS_to_PS_FSQuad input) : SV_Target
{
    float4 source;
    float2 texelIncrement = float2(0,1.0 / g_viewportSize.y);
    source  = g_HDR.Sample(g_samAniso,input.tex.xy +   texelIncrement) * (WT9_1/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy + 2*texelIncrement) * (WT9_2/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy + 3*texelIncrement) * (WT9_3/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy + 4*texelIncrement) * (WT9_4/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy                   ) * (WT9_0/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy -   texelIncrement) * (WT9_1/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy - 2*texelIncrement) * (WT9_2/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy - 3*texelIncrement) * (WT9_3/WT9_NORMALIZE);
    source += g_HDR.Sample(g_samAniso,input.tex.xy - 4*texelIncrement) * (WT9_4/WT9_NORMALIZE);
    return source;
}

//--------------------------------------------------------------------------------------
// Renders scene to Render target using D3D10 Techniques
//--------------------------------------------------------------------------------------

technique10 RenderAnimationInstanced
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, CharacterAnimatedInstancedVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, CharacterPS() ) );
        SetBlendState(Opaque,float4(0,0,0,0),0xffffffff);
        SetDepthStencilState( EnableDepthTestWrite, 1 );
        //SetRasterizerState(Wire);
        SetRasterizerState(Cull);
        SetBlendState(Opaque,float4(0,0,0,0),0xffffffff);
    }
}

technique10 RenderAnimationInstanced_MidLOD
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, CharacterAnimatedInstancedVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, CharacterMidLODPS() ) );
        SetBlendState(Opaque,float4(0,0,0,0),0xffffffff);
        SetDepthStencilState( EnableDepthTestWrite, 1 );
        //SetRasterizerState(Wire);
        SetRasterizerState(Cull);
    }
}

technique10 RenderAnimationInstanced_FarLOD
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, CharacterAnimatedInstancedVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, CharacterLowLODPS() ) );
        SetBlendState(Opaque,float4(0,0,0,0),0xffffffff);
        SetDepthStencilState( EnableDepthTestWrite, 1 );
        SetRasterizerState(NoCull);
    }
}

technique10 RenderAnimation
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, CharacterAnimatedVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, CharacterPS() ) );
        SetDepthStencilState( EnableDepthTestWrite, 1 );
        SetBlendState(Opaque,float4(0,0,0,0),0xffffffff);
        SetRasterizerState(Cull);
    }
}

technique10 RenderAnimation_MidLOD
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, CharacterAnimatedVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, CharacterMidLODPS() ) );
        SetDepthStencilState( EnableDepthTestWrite, 1 );
        SetBlendState(Opaque,float4(0,0,0,0),0xffffffff);
        SetRasterizerState(Cull);
    }
}


technique10 RenderAnimation_FarLOD
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, CharacterAnimatedVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, CharacterLowLODPS() ) );
        SetDepthStencilState( EnableDepthTestWrite, 1 );
        SetBlendState(Opaque,float4(0,0,0,0),0xffffffff);
        SetRasterizerState(NoCull);
    }
}

technique10 RenderTerrain
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, TerrainVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, TerrainPS() ) );
        SetDepthStencilState( EnableDepthTestWrite, 1 );
        SetBlendState(Opaque,float4(0,0,0,0),0xffffffff);
        SetRasterizerState(Cull);
    }
}

technique10 RenderSceneHDRtoLDR
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, FSQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, ExposureGammaPS() ) );
        SetDepthStencilState( DisableDepthTestWrite, 1 );
        SetRasterizerState(NoCull);
    }
}

technique10 BlurH
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, FSQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurHPS() ) );
        SetDepthStencilState( DisableDepthTestWrite, 1 );
        SetRasterizerState(NoCull);
    }
}

technique10 BlurV
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, FSQuadVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, BlurVPS() ) );
        SetDepthStencilState( DisableDepthTestWrite, 1 );
        SetRasterizerState(NoCull);
    }
}

technique10 GenerateShadowMap
{
    pass P0
    {
        SetVertexShader( CompileShader( vs_4_0, CharacterAnimatedInstancedVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, WriteDepthValue() ) );
        SetDepthStencilState( EnableDepthTestWrite, 1 );
        SetRasterizerState(Cull);
        SetBlendState(ShadowMapWrite,float4(0,0,0,0),0xffffffff);
    }
}